/**
 **	File:	sql2class.cpp
 **
 **	Published / author: 2004-04-14 / grymse@alhem.net
 **/

/*
Copyright (C) 2004  Anders Hedstrom

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "sql2class.h"
#include "utils.h"
#include "gclass.h"
#include "gclasslib.h"
#include "Parse.h"
#include "list.h"

#include <mysql/mysql.h>



	TBLS *tblsbase = NULL;
	char namespc[40];
	char baseclass[40];
	short container = 0;
	bool mediumtext = false;
	bool use_stl = true;
	

std::string FIELDS::GetCType(std::string& intyp,std::string& suffix)
{
	std::string strTyp;

	suffix = "";
	// long, double, char, text, mediumtext
	if (set)
	{
		strTyp = "set_t";
		intyp = "const std::string& ";
	}
	else
	if (enum_set)
	{
		strTyp = "enum_t";
		intyp = "const std::string& ";
	}
	else
	if (!strcmp(ctype, "long"))
	{
		switch (bitsize)
		{
		case 8:
			strTyp = uns ? "unsigned char" : "char";
			break;
		case 16:
			strTyp = uns ? "unsigned short" : "short";
			break;
		case 24:
		case 32:
			strTyp = uns ? "unsigned long" : "long";
			break;
		case 64:
			strTyp = uns ? "uint64_t" : "int64_t";
			break;
		}
		intyp = strTyp;
	}
	else
	if (!strcmp(ctype, "double"))
	{
		strTyp = "double";
		intyp = strTyp;
	}
	else
	if (!strncmp(ctype, "char", 4))
	{
		if (use_stl)
		{
			strTyp = "std::string";
		}
		else
		{
			char slask[100];
			sprintf(slask,"[%d]",length + 1);
			strTyp = "char";
			suffix = slask;
		}
		intyp = "const std::string& ";
	}
	else
	if (!strcmp(ctype, "text"))
	{
		strTyp = use_stl ? "std::string" : "char *";
		intyp = "const std::string& ";
	}
	else
	if (!strcmp(ctype, "mediumtext"))
	{
		strTyp = "std::vector<std::string>";
		intyp = "std::vector<std::string>& ";
	}
	else
	{
		printf("Warning. Unsupported type '%s'\n",typ);
		intyp = "";
	}
	return strTyp;
}


bool check_index(TBLS *table,INDEX *i)
{
	for (index_v::iterator it = table -> index.begin(); it != table -> index.end(); it++)
	{
		INDEX *p = *it;
		if (p -> primary || p -> unique)
		{
			fields_v::iterator it1 = i -> fields.begin();
			fields_v::iterator it2 = p -> fields.begin();
			while (it1 < i -> fields.end() && it2 < p -> fields.end())
			{
				FIELDS *f1 = *it1;
				FIELDS *f2 = *it2;
				if (!strcmp(f1 -> ctype,"double") || !strcmp(f2 -> ctype,"double"))
				{
					if (strcmp(f1 -> ctype,f2 -> ctype))
					{
						return false;
					}
				}
				else
				if (!strcmp(f1 -> ctype,"long") || !strcmp(f2 -> ctype,"long"))
				{
					if (strcmp(f1 -> ctype,f2 -> ctype))
					{
						return false;
					}
				}
				it1++;
				it2++;
				if (it1 == i -> fields.end() && it2 == p -> fields.end())
				{
					return true;
				}
				if (it1 == i -> fields.end() || it2 == p -> fields.end())
				{
					return false;
				}
			}
		}
	}
	return true;
}


void parse_index(TBLS *table,bool primary,bool unique,char *name,char *fields)
{
	INDEX *p = new INDEX(name,primary,unique);

	Parse pa(fields,"(,)");
	char slask[200];
	pa.getword(slask);
	while (*slask)
	{
		FIELDS *f;
		if (*slask == '`')
		{
			memcpy(slask, slask + 1, strlen(slask) + 1);
			slask[strlen(slask) - 1] = 0;
		}
		for (f = table -> fieldsbase; f; f = f -> next)
		{
			if (!strcasecmp(f -> column,slask))
			{
				break;
			}
		}
		if (!f)
		{
			printf("Can't resolve column name in index '%s': '%s'\n",name,slask);
			delete p;
			return;
		}
		p -> fields.push_back(f);
		//
		pa.getword(slask);
	}
	p -> ambig = (primary || unique) ? check_index(table,p) : false;
	table -> index.push_back(p);
	if (p -> primary)
	{
		table -> primary = p;
	}
}


short inset(const char *s,const char *set)
{
	std::string str;
	char slask[200];
	{
		Parse pa( (char *)s,"()");
		pa.getword(slask);
		str = slask;
	}
	Parse pa( (char *)set);
	short r = 1;

	pa.getword(slask);
	while (*slask)
	{
		if (!strcasecmp(str.c_str(),slask))
		{
			return r;
		}
		pa.getword(slask);
		r++;
	}
	return 0;
}


FIELDS *reg_column(TBLS *table,const char *n,const char *typ,bool uns)
{
	FIELDS *t;
	short l;	// type name length
	char len[200];
	char name[200];
	strcpy(name,n);

	if (*name == '`')
	{
		memcpy(name, name + 1, strlen(name + 1) + 1);
		name[strlen(name) - 1] = 0;
	}
	strlwr(name);

	t = new FIELDS;
	t -> num = listlen(table -> fieldsbase);
	strcpy(t -> column,name);
	strcpy(t -> typ,typ);
	*t -> ctype = 0;
	t -> length = 0;
	*t -> comment = 0;
	t -> ignore = false; // wont use field in insert/update if true
	t -> uns = uns;
	t -> bitsize = 0;
	t -> enum_set = false;
	t -> set = false;


	if ((l = inset(typ,"tinyint bit bool boolean smallint mediumint int integer")) != 0)
	{
		strcpy(t -> ctype,"long");
		if (inset(typ,"tinyint bit bool boolean"))
			t -> bitsize = 8;
		else
		if (inset(typ,"smallint"))
			t -> bitsize = 16;
		else
		if (inset(typ,"mediumint"))
			t -> bitsize = 24;
		else
		if (inset(typ,"int integer"))
			t -> bitsize = 32;
	}
	else
	if (!strncasecmp(typ,"bigint",6))
	{
		strcpy(t -> ctype,"long");
		t -> bitsize = 64;
	} else
	if (inset(typ,"float double double_precision real decimal dec numeric fixed")) //!strncasecmp(typ,"decimal",7) || !strncasecmp(typ,"float",5))
	{
		strcpy(t -> ctype,"double");
	} else
	if (!strncasecmp(typ,"timestamp",9))
	{
		strcpy(t -> ctype,"char[40]");
		t -> length = 40;
		t -> ignore = true; // use timestamp for timestamps
	}
	else
	if (inset(typ,"date datetime time")) //!strcasecmp(typ,"datetime") || !strcasecmp(typ,"date") || !strcasecmp(typ,"time") || !strncasecmp(typ,"timestamp",9) )
	{
		strcpy(t -> ctype,"char[40]");
		t -> length = 40;
	} else
	if (!strncasecmp(typ,"year",4))
	{
		strcpy(t -> ctype,"long");
		t -> uns = true;
		t -> bitsize = 16;
	} else
	// string types
	if (!strncasecmp(typ,"char",4))
	{
		Parse pa( (char *)typ,"()");
		pa.getword();
		pa.getword(len);
		if (!*len)
			strcpy(len,"1");
		t -> length = atoi(len); // "0" allowed
		sprintf(t -> ctype,"char[%d]",t -> length + 1);
	} else
	if (!strncasecmp(typ,"varchar",7))
	{
		Parse pa( (char *)typ,"()");
		pa.getword(); // type name
		t -> length = pa.getvalue();
		sprintf(t -> ctype,"char[%d]",t -> length + 1);
	} else
	if (!strcasecmp(typ,"tinytext"))
	{
		t -> length = 255;
		strcpy(t -> ctype,"char[256]");
	}
	else
	if (!strcasecmp(typ,"text"))
	{
		strcpy(t -> ctype,"text");
		t -> length = -1;
	} else
	if (!strcasecmp(typ,"mediumtext") || !strcasecmp(typ,"longtext") )
	{
		strcpy(t -> ctype,"mediumtext");
		t -> length = -1;
//		t -> ignore = true;
		mediumtext = true;
	} else
/*
  color enum('Red','Green','Blue') default NULL,
  colormix set('Red','Green','Blue') default NULL,
*/
	if (!strncasecmp(typ,"enum",4)) // 65535 values
	{
/*
		strcpy(t -> ctype, "long");
		t -> uns = true;
		t -> bitsize = 16;
*/
		t -> enum_set = true;
		strcpy(t -> ctype, "char[256]");
		t -> length = 255;
		//
		Parse pa( (char *)typ,"(',)");
		char slask[200];
		pa.getword(); // 'enum'
		pa.getword(slask);
		while (*slask)
		{
			t -> mvec.push_back( slask );
			//
			pa.getword(slask);
		}
	} else
	if (!strncasecmp(typ,"set",3)) // 64 bits
	{
/*
		strcpy(t -> ctype, "long");
		t -> uns = true;
		t -> bitsize = 64;
*/
		t -> enum_set = true;
		t -> set = true;
		strcpy(t -> ctype, "char[256]");
		t -> length = 255;
		//
		Parse pa( (char *)typ,"(',)");
		char slask[200];
		pa.getword(); // 'set'
		pa.getword(slask);
		while (*slask)
		{
			t -> mvec.push_back( slask );
			//
			pa.getword(slask);
		}
	} else
	{
		printf("Warning. Unsupported column data type: %s.%s '%s'\n",table -> name,name,typ);
		*t -> ctype = 0;
		t -> length = 0;
		t -> ignore = true; // seems like a good idea
	}
	addlist(&table -> fieldsbase,t);
	return t;
}


/*
 *
 */

int main(int argc,char *argv[])
{
	FILE *fil;
	TBLS *table = NULL;
	FIELDS *current = NULL;
	Parse *pa;
	int loop;
	short state = 0;
	short makerel = 0;
	short overwrite = 0;
	short global = 1;
	short build = 0;
	bool primary = false;
	bool unique = false;
	bool uns = false;
	char slask[200];
	char current_column[200];
	char filename[200];
	char dbname[200];
	char includedir[200];
	char comment[200];
	char libname[200];
	char prefix[200];
	char fields[1000];
	char index_name[100];
	std::string current_type;

	strcpy(prefix,"/usr/devel");
	*filename = 0;
	*libname = 0;
	*dbname = 0;
	*namespc = 0;
	*baseclass = 0;

	for (loop = 1; loop < argc; loop++)
		if (!strcmp(argv[loop],"-container"))
			container++;
		else
		if (!strcmp(argv[loop],"-makerel"))
			makerel++;
		else
		if (!strcmp(argv[loop],"-overwrite"))
			overwrite++;
		else
		if (!strcmp(argv[loop],"-prefix") && loop < argc - 1)
		{
			loop++;
			strcpy(prefix,argv[loop]);
			if (prefix[strlen(prefix) - 1] == '/')
				prefix[strlen(prefix) - 1] = 0;
		} else
		if (!strcmp(argv[loop],"-database") && loop < argc - 1)
		{
			loop++;
			strcpy(dbname,argv[loop]);
			strlwr(dbname);
			*dbname -= 32;
		} else
		if (!strcmp(argv[loop],"-global"))
		{
			printf("-global IS the default already\n");
			global++;
		}
		else
		if (!strcmp(argv[loop],"-global-"))
			global = 0;
		else
		if (*argv[loop] != '-')
			strcpy(filename,argv[loop]);
		else
		if (!strcmp(argv[loop],"-lib") && loop < argc - 1)
		{
			loop++;
			strcpy(libname,argv[loop]);
		} else
		if (!strcmp(argv[loop],"-build"))
			build++;
		else
		if (!strcmp(argv[loop],"-namespace") && loop < argc - 1)
		{
			loop++;
			strcpy(namespc,argv[loop]);
		}
		else
		if (!strcmp(argv[loop],"-baseclass") && loop < argc - 1)
		{
			loop++;
			strcpy(baseclass,argv[loop]);
		}
		else
		if (!strcmp(argv[loop],"-stl"))
		{
			printf("-stl IS the default already\n");
			use_stl = true;
		}
		else
		if (!strcmp(argv[loop],"-stl-"))
			use_stl = false;
		else
		{
			printf("Unknown option '%s'...\n",argv[loop]);
		}
	sprintf(includedir,"%s/include",prefix);

	if (argc < 2)
	{
		printf("Usage: %s [options] <.sql file>\n",*argv);
		printf(" options:\n");
		printf("  -prefix <path>set global directory (default '/usr/devel')\n");
		printf("  -build	build lib if '-global' and '-lib' is used\n");
		printf("  -global	generate .h files to global include directory (default)\n");
		printf("		 (default /usr/devel/include/)\n");
		printf("  -global-	turn off global generation\n");
		printf("  -stl          use std::string in generated classes (default)\n");
		printf("  -stl-         do not use STL in generated classes\n");
		printf("  -lib <name>	generate class library\n");
		printf("  -namespace <name> use namespace\n");
		printf("  -overwrite	existing files are overwritten\n");
//		printf("  -makerel	make relation class for three input tables\n");
//		printf("		 (tbl1-1, tbl2-n, tbl3-link table)\n");
		printf("  -database <name>enter database name in generated file names\n");
		printf("		 (sql<Databasename>Tbl<table>.h)\n");
		printf("  -baseclass <name> generated classes inherits from...\n");
		printf("  -container	generate old style container class also\n");
		exit(-1);
	}
	if (!*filename)
		fil = stdin;
	else
		if ((fil = fopen(filename,"rt")) == NULL)
		{
			printf("%s: Couldn't open '%s'...\n",*argv,argv[1]);
			exit(-1);
		}


// read tables

	fgets(slask,200,fil);
	while (!feof(fil))
	{
		slask[strlen(slask) - 1] = 0;
		pa = new Parse(slask);
		pa -> setbreak(',');
		pa -> disablebreak('(');
		pa -> enablebreak(')');
		pa -> getword(slask);
		while (*slask)
		{
			switch (state)
			{
			case 0:		// wait for 'create'
				if (!strcasecmp(slask,"create"))
					state = 1;
				break;
			case 1:		// wait for 'table'
				if (!strcasecmp(slask,"table"))
					state = 2;
				else
				{
					current = NULL;
					state = 0;
				}
				break;
			case 2:		// get table name
				table = new TBLS;
				if (*slask == '`')
				{
					memcpy(slask, slask + 1, strlen(slask) + 1);
					slask[strlen(slask) - 1] = 0;
				}
				strcpy(table -> name,slask);
				table -> fieldsbase = NULL;
				table -> primary = NULL;
				addlist(&tblsbase,table);

				state = 3;
				break;
			case 3:		// get '('
				if (!strcmp(slask,"("))
					state = 4;
				else
				{
					current = NULL;
					state = 0;
				}
				break;
			case 4:		// get column name or ')' end of table
				if (!strncmp(slask,"/*",2))
				{
					*comment = 0;
					state = 7;
				} else
				if (!strcasecmp(slask,"primary") ||
				    !strcasecmp(slask,"key") ||
				    !strcasecmp(slask,"unique") ||
				    !strcasecmp(slask,"index"))
				{
				// TODO: use index to build queries?
					primary = !strcasecmp(slask,"primary") ? true : false;
					unique = !strcasecmp(slask,"unique") ? true : false;
					*index_name = 0;
					*fields = 0;
					state = 8;
				} else
				if (*slask == ')' || !strcmp(slask,");"))
				{
					current = NULL;
					state = 0;
				} else
				{
					strcpy(current_column,slask);
					current_type = "";
					uns = false;
					state = 5;
				}
				break;
			case 5:		// get column type
				current_type = slask;
				state = 51;
				break;
			case 51:		// wait for ',' or ')'
				if (*slask == ',' || *slask == ')')
				{
					current = reg_column(table,current_column,current_type.c_str(),uns);
					if (*slask == ',')
						state = 4;
					else
					if (*slask == ')')
					{
						current = NULL;
						state = 0;
					}
				}
				else
				if (!strcasecmp(slask,"unsigned"))
					uns = true;
				break;
			case 6:		// wait for ',' or ')'
				if (*slask == ',')
					state = 4;
				else
				if (*slask == ')')
				{
					current = NULL;
					state = 0;
				}
				break;
			case 7:
				while (strcmp(slask + strlen(slask) - 2,"*/"))
				{
					if (*comment)
						strcat(comment," ");
					strcat(comment,slask);
					pa -> getword(slask);
				}
//				printf("comment: '%s'<br>\n",comment);
				if (current)
					strcpy(current -> comment,comment);
				state = 4;
				break;
			case 8: // index
				if (!strcasecmp(slask,"primary"))
					primary = true;
				else
				if (!strcasecmp(slask,"unique"))
					unique = true;
				else
				if (!strcasecmp(slask,"key") || !strcasecmp(slask,"index"))
				{
				}
				else
				if (*slask == '(') // fields
				{
					strcpy(fields,slask);
					if (slask[strlen(slask) - 1] == ')')
					{
						parse_index(table,primary,unique,index_name,fields);
						state = 6;
					}
					else
						state = 9;
				}
				else // index name
				{
					strcpy(index_name, slask);
				}
				break;
			case 9: // get index fields
				strcat(fields, slask);
				if (fields[strlen(fields) - 1] == ')')
				{
					parse_index(table,primary,unique,index_name,fields);
					state = 6;
				}
				break;
			}
			//
			pa -> getword(slask);
		}
		delete pa;
		fgets(slask,200,fil);
	}
	fclose(fil);


// generate source

	if (makerel)
	{
		if (listlen(tblsbase) != 3)
		{
			fprintf(stderr,"You need exactly three tables to generate this.\n");
			exit(-1);
		}
		// tbl3: [num] [ptrtbl1] [ptrtbl2] [valuex] [valuey] [charz]

//		makerelclass(tblsbase,dbname);
	} else
	if (*libname)
	{
		generate_lib(overwrite,dbname,global,includedir,libname,prefix);
		if (build && !global)
			printf("Won't build if it's not -global\n");
		if (build && global)
		{
			sprintf(slask,"%s/src/%s",prefix,libname);
			if (!chdir(slask))
				system("make install");
			else
				printf("chdir('%s') failed\n",slask);
		}
	} else
	{
		printf("Warning. Using obsolete code generator - please use '-lib' option\n");
		for (table = tblsbase; table; table = table -> next)
			generate_class(table,overwrite,dbname,global,includedir);
	}
}

